package net.knarcraft.knargui.item;

import org.bukkit.enchantments.Enchantment;
import org.bukkit.inventory.ItemFlag;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.ItemMeta;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.function.Consumer;

/**
 * An abstract factory for creating GUI items
 *
 * @param <B> <p>A child class implementing this class</p>
 */
@SuppressWarnings("unused")
public abstract class AbstractGUIItemFactory<B extends AbstractGUIItemFactory<B>> implements GUIItemFactory {

    private final ItemStack itemStack;
    private B child;

    /**
     * Instantiates a new abstract GUI item factory
     *
     * @param itemStack <p>The item stack to use for this factory</p>
     */
    protected AbstractGUIItemFactory(@NotNull ItemStack itemStack) {
        this.itemStack = itemStack;
    }

    /**
     * Sets the instance of the child class extending this class
     *
     * <p>This makes the factory return the given object. Setting this allows the factory to return a specific
     * implementation every time, without using an unsafe cast.</p>
     *
     * @param child <p>The child class extending this</p>
     */
    protected void setChild(@NotNull B child) {
        this.child = child;
    }

    @Override
    @NotNull
    public ItemStack build() {
        return this.itemStack;
    }

    @Override
    @NotNull
    public B setName(@NotNull String name) {
        return changeItemMeta((meta) -> meta.setDisplayName(name));
    }

    @Override
    @NotNull
    public B setLore(@NotNull List<String> lore) {
        return changeItemMeta((meta) -> meta.setLore(lore));
    }

    @Override
    @NotNull
    public B setEnchanted(boolean enchanted) {
        if (enchanted) {
            changeItemMeta((meta) -> meta.addEnchant(Enchantment.MENDING, 1, true));
            return changeItemMeta((meta) -> meta.addItemFlags(ItemFlag.HIDE_ENCHANTS));
        } else {
            //Remove all enchantments
            changeItemMeta((meta) -> meta.removeEnchant(Enchantment.MENDING));
            return changeItemMeta((meta) -> meta.removeItemFlags(ItemFlag.HIDE_ENCHANTS));
        }
    }

    @Override
    @NotNull
    public B changeItemMeta(@NotNull Consumer<ItemMeta> action) {
        return changeItemMeta(ItemMeta.class, action);
    }

    @Override
    @NotNull
    public <M extends ItemMeta> B changeItemMeta(@NotNull Class<M> metaType, @NotNull Consumer<M> action) {
        M meta = checkForNullMetadata(metaType);
        action.accept(meta);
        this.itemStack.setItemMeta(meta);
        if (child == null) {
            throw new IllegalStateException("Child class instance is not defined.");
        }
        return child;
    }

    @Override
    @NotNull
    public <M extends ItemMeta> M checkForNullMetadata(@NotNull Class<M> metaType) {
        M meta = metaType.cast(this.itemStack.getItemMeta());
        if (meta == null) {
            throw new IllegalArgumentException("Cannot change missing metadata");
        } else {
            return meta;
        }
    }

}
